# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice
from digi.xbee.io import IOLine, IOMode

# TODO: Replace with the serial port where your local module is connected to.
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600

REMOTE_NODE_ID = "REMOTE"

DIGITAL_LINE = IOLine.DIO3_AD3
ANALOG_LINE = IOLine.DIO2_AD2

IO_SAMPLING_RATE = 5  # 5 seconds.


def main():
    print(" +----------------------------------------------+")
    print(" | XBee Python Library Handle IO Samples Sample |")
    print(" +----------------------------------------------+\n")

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        # Obtain the remote XBee device from the XBee network.
        xbee_network = device.get_network()
        remote_device = xbee_network.discover_device(REMOTE_NODE_ID)
        if remote_device is None:
            print("Could not find the remote device")
            exit(1)

        # Set the local device as destination address of the remote.
        remote_device.set_dest_address(device.get_64bit_addr())

        remote_device.set_io_configuration(DIGITAL_LINE, IOMode.DIGITAL_IN)
        remote_device.set_io_configuration(ANALOG_LINE, IOMode.ADC)

        # Enable periodic sampling every IO_SAMPLING_RATE seconds in the remote device.
        remote_device.set_io_sampling_rate(IO_SAMPLING_RATE)

        # Enable DIO change detection in the remote device.
        remote_device.set_dio_change_detection({DIGITAL_LINE})

        # Register a listener to handle the samples received by the local device.
        def io_samples_callback(sample, remote, time):
            print("New sample received from %s - %s" % (remote.get_64bit_addr(), sample))

        device.add_io_sample_received_callback(io_samples_callback)

        input()

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.util import utils


@unique
class OperatingMode(Enum):
    """
    This class represents all operating modes available.

    | Inherited properties:
    |     **name** (String): the name (id) of this OperatingMode.
    |     **value** (String): the value of this OperatingMode.
    """

    AT_MODE = (0, "AT mode")
    API_MODE = (1, "API mode")
    ESCAPED_API_MODE = (2, "API mode with escaped characters")
    UNKNOWN = (99, "Unknown")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the OperatingMode element.

        Returns:
            String: the code of the OperatingMode element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the OperatingMode element.

        Returns:
            String: the description of the OperatingMode element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the OperatingMode for the given code.

        Args:
            code (Integer): the code corresponding to the operating mode to get.

        Returns:
            :class:`.OperatingMode`: the OperatingMode with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return OperatingMode.UNKNOWN

    code = property(__get_code)
    """Integer. The operating mode code."""

    description = property(__get_description)
    """String: The operating mode description."""


OperatingMode.lookupTable = {x.code: x for x in OperatingMode}
OperatingMode.__doc__ += utils.doc_enum(OperatingMode)


@unique
class APIOutputMode(Enum):
    """
    Enumerates the different API output modes. The API output mode establishes
    the way data will be output through the serial interface of an XBee device.

    | Inherited properties:
    |     **name** (String): the name (id) of this OperatingMode.
    |     **value** (String): the value of this OperatingMode.
    """

    NATIVE = (0x00, "Native")
    EXPLICIT = (0x01, "Explicit")
    EXPLICIT_ZDO_PASSTHRU = (0x03, "Explicit with ZDO Passthru")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the APIOutputMode element.

        Returns:
            String: the code of the APIOutputMode element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the APIOutputMode element.

        Returns:
            String: the description of the APIOutputMode element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the APIOutputMode for the given code.

        Args:
            code (Integer): the code corresponding to the API output mode to get.

        Returns:
            :class:`.OperatingMode`: the APIOutputMode with the given code, ``None`` if there is not an
                APIOutputMode with that code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The API output mode code."""

    description = property(__get_description)
    """String: The API output mode description."""


APIOutputMode.lookupTable = {x.code: x for x in APIOutputMode}
APIOutputMode.__doc__ += utils.doc_enum(APIOutputMode)


@unique
class IPAddressingMode(Enum):
    """
    Enumerates the different IP addressing modes.
    """

    DHCP = (0x00, "DHCP")
    STATIC = (0x01, "Static")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the IPAddressingMode element.

        Returns:
            String: the code of the IPAddressingMode element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the IPAddressingMode element.

        Returns:
            String: the description of the IPAddressingMode element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the IPAddressingMode for the given code.

        Args:
            code (Integer): the code corresponding to the IP addressing mode to get.

        Returns:
            :class:`.IPAddressingMode`: the IPAddressingMode with the given code, ``None`` if there is not an
                IPAddressingMode with that code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The IP addressing mode code."""

    description = property(__get_description)
    """String. The IP addressing mode description."""


IPAddressingMode.lookupTable = {x.code: x for x in IPAddressingMode}
IPAddressingMode.__doc__ += utils.doc_enum(IPAddressingMode)



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.util import utils


class AccessPoint(object):
    """
    This class represents an Access Point for the Wi-Fi protocol. It contains 
    SSID, the encryption type and the link quality between the Wi-Fi module and 
    the access point.
    
    This class is used within the library to list the access points
    and connect to a specific one in the Wi-Fi protocol.

    .. seealso::
       | :class:`.WiFiEncryptionType`
    """

    __ERROR_CHANNEL = "Channel cannot be negative."
    __ERROR_SIGNAL_QUALITY = "Signal quality must be between 0 and 100."

    def __init__(self, ssid, encryption_type, channel=0, signal_quality=0):
        """
        Class constructor. Instantiates a new :class:`.AccessPoint` object with the provided parameters.

        Args:
            ssid (String): the SSID of the access point.
            encryption_type (:class:`.WiFiEncryptionType`): the encryption type configured in the access point.
            channel (Integer, optional): operating channel of the access point. Optional.
            signal_quality (Integer, optional): signal quality with the access point in %. Optional.

        Raises:
            ValueError: if length of ``ssid`` is 0.
            ValueError: if ``channel`` is less than 0.
            ValueError: if ``signal_quality`` is less than 0 or greater than 100.

        .. seealso::
           | :class:`.WiFiEncryptionType`
        """
        if len(ssid) == 0:
            raise ValueError("SSID cannot be empty.")
        if channel < 0:
            raise ValueError(self.__ERROR_CHANNEL)
        if signal_quality < 0 or signal_quality > 100:
            raise ValueError(self.__ERROR_SIGNAL_QUALITY)

        self.__ssid = ssid
        self.__encryption_type = encryption_type
        self.__channel = channel
        self.__signal_quality = signal_quality

    def __str__(self):
        """
        Returns the string representation of the access point.

        Returns:
            String: representation of the access point.
        """
        return "%s (%s) - CH: %s - Signal: %s%%" % (self.__ssid, self.__encryption_type.description,
                                                    self.__channel, self.__signal_quality)

    def __get_ssid(self):
        """
        Returns the SSID of the access point.

        Returns:
            String: the SSID of the access point.
        """
        return self.__ssid

    def __get_encryption_type(self):
        """
        Returns the encryption type of the access point.

        Returns:
            :class:`.WiFiEncryptionType`: the encryption type of the access point.

        .. seealso::
           | :class:`.WiFiEncryptionType`
        """
        return self.__encryption_type

    def __get_channel(self):
        """
        Returns the channel of the access point.

        Returns:
            Integer: the channel of the access point.

        .. seealso::
           | :func:`.AccessPoint.set_channel`
        """
        return self.__channel

    def __set_channel(self, channel):
        """
        Sets the channel of the access point.

        Args:
            channel (Integer): the new channel of the access point

        Raises:
            ValueError: if ``channel`` is less than 0.

        .. seealso::
           | :func:`.AccessPoint.get_channel`
        """
        if channel < 0:
            raise ValueError(self.__ERROR_CHANNEL)
        self.__channel = channel

    def __get_signal_quality(self):
        """
        Returns the signal quality with the access point in %.

        Returns:
            Integer: the signal quality with the access point in %.

        .. seealso::
           | :func:`.AccessPoint.__set_signal_quality`
        """
        return self.__signal_quality

    def __set_signal_quality(self, signal_quality):
        """
        Sets the channel of the access point.

        Args:
            signal_quality (Integer): the new signal quality with the access point.

        Raises:
            ValueError: if ``signal_quality`` is less than 0 or greater than 100.

        .. seealso::
           | :func:`.AccessPoint.__get_signal_quality`
        """
        if signal_quality < 0 or signal_quality > 100:
            raise ValueError(self.__ERROR_SIGNAL_QUALITY)
        self.__signal_quality = signal_quality

    ssid = property(__get_ssid)
    """String. SSID of the access point."""

    encryption_type = property(__get_encryption_type)
    """:class:`.WiFiEncryptionType`. Encryption type of the access point."""

    channel = property(__get_channel, __set_channel)
    """String. Channel of the access point."""

    signal_quality = property(__get_signal_quality, __set_signal_quality)
    """String. The signal quality with the access point in %."""


@unique
class WiFiEncryptionType(Enum):
    """
    Enumerates the different Wi-Fi encryption types.
    """
    NONE = (0, "No security")
    WPA = (1, "WPA (TKIP) security")
    WPA2 = (2, "WPA2 (AES) security")
    WEP = (3, "WEP security")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the WiFiEncryptionType element.

        Returns:
            Integer: the code of the WiFiEncryptionType element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the WiFiEncryptionType element.

        Returns:
            String: the description of the WiFiEncryptionType element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the Wi-Fi encryption type for the given code.

        Args:
            code (Integer): the code of the Wi-Fi encryption type to get.

        Returns:
            :class:`.WiFiEncryptionType`: the WiFiEncryptionType with the given code, ``None`` if there is
                not any Wi-Fi encryption type with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The Wi-Fi encryption type code."""

    description = property(__get_description)
    """String. The Wi-Fi encryption type description."""


WiFiEncryptionType.lookupTable = {x.code: x for x in WiFiEncryptionType}
WiFiEncryptionType.__doc__ += utils.doc_enum(WiFiEncryptionType)



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.util import utils


@unique
class ATCommandStatus(Enum):
    """
    This class lists all the possible states of an AT command after executing it.

    | Inherited properties:
    |     **name** (String): the name (id) of the ATCommandStatus.
    |     **value** (String): the value of the ATCommandStatus.
    """
    OK = (0, "Status OK")
    ERROR = (1, "Status Error")
    INVALID_COMMAND = (2, "Invalid command")
    INVALID_PARAMETER = (3, "Invalid parameter")
    TX_FAILURE = (4, "TX failure")
    UNKNOWN = (255, "Unknown status")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ATCommandStatus element.

        Returns:
            Integer: the code of the ATCommandStatus element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ATCommandStatus element.

        Returns:
            String: the description of the ATCommandStatus element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the AT command status for the given code.

        Args:
            code (Integer): the code of the AT command status to get.

        Returns:
            :class:`.ATCommandStatus`: the AT command status with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return ATCommandStatus.UNKNOWN

    code = property(__get_code)
    """Integer. The AT command status code."""

    description = property(__get_description)
    """String. The AT command status description."""


ATCommandStatus.lookupTable = {x.code: x for x in ATCommandStatus}
ATCommandStatus.__doc__ += utils.doc_enum(ATCommandStatus)


@unique
class DiscoveryStatus(Enum):
    """
    This class lists all the possible states of the discovery process.

    | Inherited properties:
    |     **name** (String): The name of the DiscoveryStatus.
    |     **value** (Integer): The ID of the DiscoveryStatus.
    """
    NO_DISCOVERY_OVERHEAD = (0x00, "No discovery overhead")
    ADDRESS_DISCOVERY = (0x01, "Address discovery")
    ROUTE_DISCOVERY = (0x02, "Route discovery")
    ADDRESS_AND_ROUTE = (0x03, "Address and route")
    EXTENDED_TIMEOUT_DISCOVERY = (0x40, "Extended timeout discovery")
    UNKNOWN = (0xFF, "Unknown")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the DiscoveryStatus element.

        Returns:
            Integer: the code of the DiscoveryStatus element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the DiscoveryStatus element.

        Returns:
            String: The description of the DiscoveryStatus element.

        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the discovery status for the given code.

        Args:
            code (Integer): the code of the discovery status to get.

        Returns:
            :class:`.DiscoveryStatus`: the discovery status with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return DiscoveryStatus.UNKNOWN

    code = property(__get_code)
    """Integer. The discovery status code."""

    description = property(__get_description)
    """String. The discovery status description."""


DiscoveryStatus.lookupTable = {x.code: x for x in DiscoveryStatus}
DiscoveryStatus.__doc__ += utils.doc_enum(DiscoveryStatus)


@unique
class TransmitStatus(Enum):
    """
    This class represents all available transmit status.

    | Inherited properties:
    |     **name** (String): the name (id) of ths TransmitStatus.
    |     **value** (String): the value of ths TransmitStatus.
    """
    SUCCESS = (0x00, "Success.")
    NO_ACK = (0x01, "No acknowledgement received.")
    CCA_FAILURE = (0x02, "CCA failure.")
    PURGED = (0x03, "Transmission purged, it was attempted before stack was up.")
    WIFI_PHYSICAL_ERROR = (0x04, "Physical error occurred on the interface with the WiFi transceiver.")
    INVALID_DESTINATION = (0x15, "Invalid destination endpoint.")
    NO_BUFFERS = (0x18, "No buffers.")
    NETWORK_ACK_FAILURE = (0x21, "Network ACK Failure.")
    NOT_JOINED_NETWORK = (0x22, "Not joined to network.")
    SELF_ADDRESSED = (0x23, "Self-addressed.")
    ADDRESS_NOT_FOUND = (0x24, "Address not found.")
    ROUTE_NOT_FOUND = (0x25, "Route not found.")
    BROADCAST_FAILED = (0x26, "Broadcast source failed to hear a neighbor relay the message.")
    INVALID_BINDING_TABLE_INDEX = (0x2B, "Invalid binding table index.")
    INVALID_ENDPOINT = (0x2C, "Invalid endpoint")
    BROADCAST_ERROR_APS = (0x2D, "Attempted broadcast with APS transmission.")
    BROADCAST_ERROR_APS_EE0 = (0x2E, "Attempted broadcast with APS transmission, but EE=0.")
    SOFTWARE_ERROR = (0x31, "A software error occurred.")
    RESOURCE_ERROR = (0x32, "Resource error lack of free buffers, timers, etc.")
    PAYLOAD_TOO_LARGE = (0x74, "Data payload too large.")
    INDIRECT_MESSAGE_UNREQUESTED = (0x75, "Indirect message unrequested")
    SOCKET_CREATION_FAILED = (0x76, "Attempt to create a client socket failed.")
    IP_PORT_NOT_EXIST = (0x77, "TCP connection to given IP address and port doesn't exist. Source port is non-zero so "
                               "that a new connection is not attempted.")
    UDP_SRC_PORT_NOT_MATCH_LISTENING_PORT = (0x78, "Source port on a UDP transmission doesn't match a listening port "
                                                   "on the transmitting module.")
    KEY_NOT_AUTHORIZED = (0xBB, "Key not authorized.")
    UNKNOWN = (0xFF, "Unknown.")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the TransmitStatus element.

        Returns:
            Integer: the code of the TransmitStatus element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the TransmitStatus element.

        Returns:
            String: the description of the TransmitStatus element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the transmit status for the given code.

        Args:
            code (Integer): the code of the transmit status to get.

        Returns:
            :class:`.TransmitStatus`: the transmit status with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return TransmitStatus.UNKNOWN

    code = property(__get_code)
    """Integer. The transmit status code."""

    description = property(__get_description)
    """String. The transmit status description."""


TransmitStatus.lookupTable = {x.code: x for x in TransmitStatus}
TransmitStatus.__doc__ += utils.doc_enum(TransmitStatus)


@unique
class ModemStatus(Enum):
    """
    Enumerates the different modem status events. This enumeration list is 
    intended to be used within the :class:`.ModemStatusPacket` packet.
    """
    HARDWARE_RESET = (0x00, "Device was reset")
    WATCHDOG_TIMER_RESET = (0x01, "Watchdog timer was reset")
    JOINED_NETWORK = (0x02, "Device joined to network")
    DISASSOCIATED = (0x03, "Device disassociated")
    ERROR_SYNCHRONIZATION_LOST = (0x04, "Configuration error/synchronization lost")
    COORDINATOR_REALIGNMENT = (0x05, "Coordinator realignment")
    COORDINATOR_STARTED = (0x06, "The coordinator started")
    NETWORK_SECURITY_KEY_UPDATED = (0x07, "Network security key was updated")
    NETWORK_WOKE_UP = (0x0B, "Network Woke Up")
    NETWORK_WENT_TO_SLEEP = (0x0C, "Network Went To Sleep")
    VOLTAGE_SUPPLY_LIMIT_EXCEEDED = (0x0D, "Voltage supply limit exceeded")
    REMOTE_MANAGER_CONNECTED = (0x0E, "Remote Manager connected")
    REMOTE_MANAGER_DISCONNECTED = (0x0F, "Remote Manager disconnected")
    MODEM_CONFIG_CHANGED_WHILE_JOINING = (0x11, "Modem configuration changed while joining")
    ERROR_STACK = (0x80, "Stack error")
    ERROR_AP_NOT_CONNECTED = (0x82, "Send/join command issued without connecting from AP")
    ERROR_AP_NOT_FOUND = (0x83, "Access point not found")
    ERROR_PSK_NOT_CONFIGURED = (0x84, "PSK not configured")
    ERROR_SSID_NOT_FOUND = (0x87, "SSID not found")
    ERROR_FAILED_JOIN_SECURITY = (0x88, "Failed to join with security enabled")
    ERROR_INVALID_CHANNEL = (0x8A, "Invalid channel")
    ERROR_FAILED_JOIN_AP = (0x8E, "Failed to join access point")
    UNKNOWN = (0xFF, "UNKNOWN")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ModemStatus element.

        Returns:
            Integer: the code of the ModemStatus element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ModemStatus element.

        Returns:
            String: the description of the ModemStatus element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the modem status for the given code.

        Args:
            code (Integer): the code of the modem status to get.

        Returns:
            :class:`.ModemStatus`: the ModemStatus with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return ModemStatus.UNKNOWN

    code = property(__get_code)
    """Integer. The modem status code."""

    description = property(__get_description)
    """String. The modem status description."""


ModemStatus.lookupTable = {x.code: x for x in ModemStatus}
ModemStatus.__doc__ += utils.doc_enum(ModemStatus)


@unique
class PowerLevel(Enum):
    """
    Enumerates the different power levels. The power level indicates the output 
    power value of a radio when transmitting data.
    """
    LEVEL_LOWEST = (0x00, "Lowest")
    LEVEL_LOW = (0x01, "Low")
    LEVEL_MEDIUM = (0x02, "Medium")
    LEVEL_HIGH = (0x03, "High")
    LEVEL_HIGHEST = (0x04, "Highest")
    LEVEL_UNKNOWN = (0xFF, "Unknown")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the PowerLevel element.

        Returns:
            Integer: the code of the PowerLevel element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the PowerLevel element.

        Returns:
            String: the description of the PowerLevel element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the power level for the given code.

        Args:
            code (Integer): the code of the power level to get.

        Returns:
            :class:`.PowerLevel`: the PowerLevel with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return PowerLevel.LEVEL_UNKNOWN

    code = property(__get_code)
    """Integer. The power level code."""

    description = property(__get_description)
    """String. The power level description."""


PowerLevel.lookupTable = {x.code: x for x in PowerLevel}
PowerLevel.__doc__ += utils.doc_enum(PowerLevel)


@unique
class AssociationIndicationStatus(Enum):
    """
    Enumerates the different association indication statuses.
    """
    SUCCESSFULLY_JOINED = (0x00, "Successfully formed or joined a network.")
    AS_TIMEOUT = (0x01, "Active Scan Timeout.")
    AS_NO_PANS_FOUND = (0x02, "Active Scan found no PANs.")
    AS_ASSOCIATION_NOT_ALLOWED = (0x03, "Active Scan found PAN, but the CoordinatorAllowAssociation bit is not set.")
    AS_BEACONS_NOT_SUPPORTED = (0x04, "Active Scan found PAN, but Coordinator and End Device are not configured to "
                                      "support beacons.")
    AS_ID_DOESNT_MATCH = (0x05, "Active Scan found PAN, but the Coordinator ID parameter does not match the ID "
                                "parameter of the End Device.")
    AS_CHANNEL_DOESNT_MATCH = (0x06, "Active Scan found PAN, but the Coordinator CH parameter does not match "
                                     "the CH parameter of the End Device.")
    ENERGY_SCAN_TIMEOUT = (0x07, "Energy Scan Timeout.")
    COORDINATOR_START_REQUEST_FAILED = (0x08, "Coordinator start request failed.")
    COORDINATOR_INVALID_PARAMETER = (0x09, "Coordinator could not start due to invalid parameter.")
    COORDINATOR_REALIGNMENT = (0x0A, "Coordinator Realignment is in progress.")
    AR_NOT_SENT = (0x0B, "Association Request not sent.")
    AR_TIMED_OUT = (0x0C, "Association Request timed out - no reply was received.")
    AR_INVALID_PARAMETER = (0x0D, "Association Request had an Invalid Parameter.")
    AR_CHANNEL_ACCESS_FAILURE = (0x0E, "Association Request Channel Access Failure. Request was not "
                                       "transmitted - CCA failure.")
    AR_COORDINATOR_ACK_WASNT_RECEIVED = (0x0F, "Remote Coordinator did not send an ACK after Association "
                                               "Request was sent.")
    AR_COORDINATOR_DIDNT_REPLY = (0x10, "Remote Coordinator did not reply to the Association Request, but an ACK "
                                        "was received after sending the request.")
    SYNCHRONIZATION_LOST = (0x12, "Sync-Loss - Lost synchronization with a Beaconing Coordinator.")
    DISASSOCIATED = (0x13, " Disassociated - No longer associated to Coordinator.")
    NO_PANS_FOUND = (0x21, "Scan found no PANs.")
    NO_PANS_WITH_ID_FOUND = (0x22, "Scan found no valid PANs based on current SC and ID settings.")
    NJ_EXPIRED = (0x23, "Valid Coordinator or Routers found, but they are not allowing joining (NJ expired).")
    NO_JOINABLE_BEACONS_FOUND = (0x24, "No joinable beacons were found.")
    UNEXPECTED_STATE = (0x25, "Unexpected state, node should not be attempting to join at this time.")
    JOIN_FAILED = (0x27, "Node Joining attempt failed (typically due to incompatible security settings).")
    COORDINATOR_START_FAILED = (0x2A, "Coordinator Start attempt failed.")
    CHECKING_FOR_COORDINATOR = (0x2B, "Checking for an existing coordinator.")
    NETWORK_LEAVE_FAILED = (0x2C, "Attempt to leave the network failed.")
    DEVICE_DIDNT_RESPOND = (0xAB, "Attempted to join a device that did not respond.")
    UNSECURED_KEY_RECEIVED = (0xAC, "Secure join error - network security key received unsecured.")
    KEY_NOT_RECEIVED = (0xAD, "Secure join error - network security key not received.")
    INVALID_SECURITY_KEY = (0xAF, "Secure join error - joining device does not have the right preconfigured link key.")
    SCANNING_NETWORK = (0xFF, "Scanning for a network/Attempting to associate.")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``AssociationIndicationStatus`` element.

        Returns:
            Integer: the code of the ``AssociationIndicationStatus`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``AssociationIndicationStatus`` element.

        Returns:
            String: the description of the ``AssociationIndicationStatus`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the ``AssociationIndicationStatus`` for the given code.

        Args:
            code (Integer): the code of the ``AssociationIndicationStatus`` to get.

        Returns:
            :class:`.AssociationIndicationStatus`: the ``AssociationIndicationStatus`` with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The association indication status code."""

    description = property(__get_description)
    """String. The association indication status description."""


AssociationIndicationStatus.lookupTable = {x.code: x for x in AssociationIndicationStatus}
AssociationIndicationStatus.__doc__ += utils.doc_enum(AssociationIndicationStatus)


@unique
class CellularAssociationIndicationStatus(Enum):
    """
    Enumerates the different association indication statuses for the Cellular protocol.
    """
    SUCCESSFULLY_CONNECTED = (0x00, "Connected to the Internet.")
    REGISTERING_CELLULAR_NETWORK = (0x22, "Registering to cellular network")
    CONNECTING_INTERNET = (0x23, "Connecting to the Internet")
    BYPASS_MODE = (0x2F, "Bypass mode active")
    INITIALIZING = (0xFF, "Initializing")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``CellularAssociationIndicationStatus`` element.

        Returns:
            Integer: the code of the ``CellularAssociationIndicationStatus`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``CellularAssociationIndicationStatus`` element.

        Returns:
            String: the description of the ``CellularAssociationIndicationStatus`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the ``CellularAssociationIndicationStatus`` for the given code.

        Args:
            code (Integer): the code of the ``CellularAssociationIndicationStatus`` to get.

        Returns:
            :class:`.CellularAssociationIndicationStatus`: the ``CellularAssociationIndicationStatus``
                with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The cellular association indication status code."""

    description = property(__get_description)
    """String. The cellular association indication status description."""


CellularAssociationIndicationStatus.lookupTable = {x.code: x for x in CellularAssociationIndicationStatus}
CellularAssociationIndicationStatus.__doc__ += utils.doc_enum(CellularAssociationIndicationStatus)


@unique
class DeviceCloudStatus(Enum):
    """
    Enumerates the different Device Cloud statuses.
    """
    SUCCESS = (0x00, "Success")
    BAD_REQUEST = (0x01, "Bad request")
    RESPONSE_UNAVAILABLE = (0x02, "Response unavailable")
    DEVICE_CLOUD_ERROR = (0x03, "Device Cloud error")
    CANCELED = (0x20, "Device Request canceled by user")
    TIME_OUT = (0x21, "Session timed out")
    UNKNOWN_ERROR = (0x40, "Unknown error")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``DeviceCloudStatus`` element.

        Returns:
            Integer: the code of the ``DeviceCloudStatus`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``DeviceCloudStatus`` element.

        Returns:
            String: the description of the ``DeviceCloudStatus`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the Device Cloud status for the given code.

        Args:
            code (Integer): the code of the Device Cloud status to get.

        Returns:
            :class:`.DeviceCloudStatus`: the ``DeviceCloudStatus`` with the given code, ``None`` if there is not any
                status with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The Device Cloud status code."""

    description = property(__get_description)
    """String. The Device Cloud status description."""


DeviceCloudStatus.lookupTable = {x.code: x for x in DeviceCloudStatus}
DeviceCloudStatus.__doc__ += utils.doc_enum(DeviceCloudStatus)


@unique
class FrameError(Enum):
    """
    Enumerates the different frame errors.
    """
    INVALID_TYPE = (0x02, "Invalid frame type")
    INVALID_LENGTH = (0x03, "Invalid frame length")
    INVALID_CHECKSUM = (0x04, "Erroneous checksum on last frame")
    PAYLOAD_TOO_BIG = (0x05, "Payload of last API frame was too big to fit into a buffer")
    STRING_ENTRY_TOO_BIG = (0x06, "String entry was too big on last API frame sent")
    WRONG_STATE = (0x07, "Wrong state to receive frame")
    WRONG_REQUEST_ID = (0x08, "Device request ID of device response didn't match the number in the request")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``FrameError`` element.

        Returns:
            Integer: the code of the ``FrameError`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``FrameError`` element.

        Returns:
            String: the description of the ``FrameError`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the frame error for the given code.

        Args:
            code (Integer): the code of the frame error to get.

        Returns:
            :class:`.FrameError`: the ``FrameError`` with the given code, ``None`` if there is not any frame error
                with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The frame error code."""

    description = property(__get_description)
    """String. The frame error description."""


FrameError.lookupTable = {x.code: x for x in FrameError}
FrameError.__doc__ += utils.doc_enum(FrameError)


@unique
class WiFiAssociationIndicationStatus(Enum):
    """
    Enumerates the different Wi-Fi association indication statuses.
    """
    SUCCESSFULLY_JOINED = (0x00, "Successfully joined to access point.")
    INITIALIZING = (0x01, "Initialization in progress.")
    INITIALIZED = (0x02, "Initialized, but not yet scanning.")
    DISCONNECTING = (0x13, "Disconnecting from access point.")
    SSID_NOT_CONFIGURED = (0x23, "SSID not configured")
    INVALID_KEY = (0x24, "Encryption key invalid (NULL or invalid length).")
    JOIN_FAILED = (0x27, "SSID found, but join failed.")
    WAITING_FOR_AUTH = (0x40, "Waiting for WPA or WPA2 authentication.")
    WAITING_FOR_IP = (0x41, "Joined to a network and waiting for IP address.")
    SETTING_UP_SOCKETS = (0x42, "Joined to a network and IP configured. Setting up listening sockets.")
    SCANNING_FOR_SSID = (0xFF, "Scanning for the configured SSID.")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``WiFiAssociationIndicationStatus`` element.

        Returns:
            Integer: the code of the ``WiFiAssociationIndicationStatus`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``WiFiAssociationIndicationStatus`` element.

        Returns:
            String: the description of the ``WiFiAssociationIndicationStatus`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the Wi-Fi association indication status for the given code.

        Args:
            code (Integer): the code of the Wi-Fi association indication status to get.

        Returns:
            :class:`.WiFiAssociationIndicationStatus`: the ``WiFiAssociationIndicationStatus`` with the given code,
                ``None`` if there is not any Wi-Fi association indication status with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The Wi-Fi association indication status code."""

    description = property(__get_description)
    """String. The Wi-Fi association indication status description."""


WiFiAssociationIndicationStatus.lookupTable = {x.code: x for x in WiFiAssociationIndicationStatus}
WiFiAssociationIndicationStatus.__doc__ += utils.doc_enum(WiFiAssociationIndicationStatus)


@unique
class NetworkDiscoveryStatus(Enum):
    """
    Enumerates the different statuses of the network discovery process.
    """
    SUCCESS = (0x00, "Success")
    ERROR_READ_TIMEOUT = (0x01, "Read timeout error")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``NetworkDiscoveryStatus`` element.

        Returns:
            Integer: the code of the ``NetworkDiscoveryStatus`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``NetworkDiscoveryStatus`` element.

        Returns:
            String: the description of the ``NetworkDiscoveryStatus`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the network discovery status for the given code.

        Args:
            code (Integer): the code of the network discovery status to get.

        Returns:
            :class:`.NetworkDiscoveryStatus`: the ``NetworkDiscoveryStatus`` with the given code, ``None`` if
                there is not any status with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The network discovery status code."""

    description = property(__get_description)
    """String. The network discovery status description."""


NetworkDiscoveryStatus.lookupTable = {x.code: x for x in NetworkDiscoveryStatus}
NetworkDiscoveryStatus.__doc__ += utils.doc_enum(NetworkDiscoveryStatus)



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from ipaddress import IPv4Address
from digi.xbee.models.mode import OperatingMode
from digi.xbee.models.protocol import IPProtocol
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.packets.base import XBeeAPIPacket, DictKeys
from digi.xbee.util import utils
from digi.xbee.exception import InvalidOperatingModeException, InvalidPacketException


class RXIPv4Packet(XBeeAPIPacket):
    """
    This class represents an RX (Receive) IPv4 packet. Packet is built 
    using the parameters of the constructor or providing a valid byte array.
    
    .. seealso::
       | :class:`.TXIPv4Packet`
       | :class:`.XBeeAPIPacket`
    """
    __MIN_PACKET_LENGTH = 15

    def __init__(self, source_address, dest_port, source_port, ip_protocol, data=None):
        """
        Class constructor. Instantiates a new :class:`.RXIPv4Packet` object with the provided parameters.

        Args:
            source_address (:class:`.IPv4Address`): IPv4 address of the source device.
            dest_port (Integer): destination port number.
            source_port (Integer): source port number.
            ip_protocol (:class:`.IPProtocol`): IP protocol used for transmitted data.
            data (Bytearray, optional): data that is sent to the destination device. Optional.

        Raises:
            ValueError: if ``dest_port`` is less than 0 or greater than 65535 or
            ValueError: if ``source_port`` is less than 0 or greater than 65535.

        .. seealso::
           | :class:`.IPProtocol`
        """
        if dest_port < 0 or dest_port > 65535:
            raise ValueError("Destination port must be between 0 and 65535")
        if source_port < 0 or source_port > 65535:
            raise ValueError("Source port must be between 0 and 65535")

        super().__init__(ApiFrameType.RX_IPV4)
        self.__source_address = source_address
        self.__dest_port = dest_port
        self.__source_port = source_port
        self.__ip_protocol = ip_protocol
        self.__status = 0  # Reserved
        self.__data = data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            RXIPv4Packet.

        Raises:
            InvalidPacketException: if the bytearray length is less than 15. (start delim + length (2 bytes) + frame
                type + source address (4 bytes) + dest port (2 bytes) + source port (2 bytes) + network protocol +
                status + checksum = 15 bytes)
            InvalidPacketException: if the length field of ``raw`` is different than its real length. (length field:
                bytes 2 and 3)
            InvalidPacketException: if the first byte of ``raw`` is not the header byte. See :class:`.SPECIAL_BYTE`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is not :attr:`ApiFrameType.RX_IPV4`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=RXIPv4Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_IPV4.code:
            raise InvalidPacketException("This packet is not an RXIPv4Packet.")

        return RXIPv4Packet(IPv4Address(bytes(raw[4:8])), utils.bytes_to_int(raw[8:10]),
                            utils.bytes_to_int(raw[10:12]), IPProtocol.get(raw[12]),
                            raw[14:-1])

    def needs_id(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def __get_source_address(self):
        """
        Returns the IPv4 address of the source device.

        Returns:
            :class:`ipaddress.IPv4Address`: the IPv4 address of the source device.
        """
        return self.__source_address

    def __set_source_address(self, source_address):
        """
        Sets the IPv4 source address.

        Args:
            source_address (:class:`.IPv4Address`): The new IPv4 source address.
        """
        if source_address is not None:
            self.__source_address = source_address

    def __get_dest_port(self):
        """
        Returns the destination port.

        Returns:
            Integer: the destination port.
        """
        return self.__dest_port

    def __set_dest_port(self, dest_port):
        """
        Sets the destination port.

        Args:
            dest_port (Integer): the new destination port.

        Raises:
            ValueError: if ``dest_port`` is less than 0 or greater than 65535.
        """
        if dest_port < 0 or dest_port > 65535:
            raise ValueError("Destination port must be between 0 and 65535")
        self.__dest_port = dest_port

    def __get_source_port(self):
        """
        Returns the source port.

        Returns:
            Integer: the source port.
        """
        return self.__source_port

    def __set_source_port(self, source_port):
        """
        Sets the source port.

        Args:
            source_port (Integer): the new source port.

        Raises:
            ValueError: if ``source_port`` is less than 0 or greater than 65535.
        """
        if source_port < 0 or source_port > 65535:
            raise ValueError("Source port must be between 0 and 65535")
        self.__source_port = source_port

    def __get_ip_protocol(self):
        """
        Returns the IP protocol used for transmitted data.

        Returns:
            :class:`.IPProtocol`: the IP protocol used for transmitted data.
        """
        return self.__ip_protocol

    def __set_ip_protocol(self, ip_protocol):
        """
        Sets the IP protocol used for transmitted data.

        Args:
            ip_protocol (:class:`.IPProtocol`): the new IP protocol.
        """
        self.__ip_protocol = ip_protocol

    def __get_data(self):
        """
        Returns the data of the packet.

        Returns:
            Bytearray: the data of the packet.
        """
        if self.__data is None:
            return self.__data
        return self.__data.copy()

    def __set_data(self, data):
        """
        Sets the data of the packet.

        Args:
            data (Bytearray): the new data of the packet.
        """
        if data is None:
            self.__data = None
        else:
            self.__data = data.copy()

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data`
        """
        ret = bytearray(self.__source_address.packed)
        ret += utils.int_to_bytes(self.__dest_port, num_bytes=2)
        ret += utils.int_to_bytes(self.__source_port, num_bytes=2)
        ret += utils.int_to_bytes(self.__ip_protocol.code, num_bytes=1)
        ret += utils.int_to_bytes(self.__status, num_bytes=1)
        if self.__data is not None:
            ret += self.__data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data_dict`
        """
        return {DictKeys.SRC_IPV4_ADDR: "%s (%s)" % (self.__source_address.packed, self.__source_address.exploded),
                DictKeys.DEST_PORT:     self.__dest_port,
                DictKeys.SRC_PORT:      self.__source_port,
                DictKeys.IP_PROTOCOL:   "%s (%s)" % (self.__ip_protocol.code, self.__ip_protocol.description),
                DictKeys.STATUS:        self.__status,
                DictKeys.RF_DATA:       bytearray(self.__data)}

    source_address = property(__get_source_address, __set_source_address)
    """:class:`ipaddress.IPv4Address`. IPv4 address of the source device."""

    dest_port = property(__get_dest_port, __set_dest_port)
    """Integer. Destination port."""

    source_port = property(__get_source_port, __set_source_port)
    """Integer. Source port."""

    ip_protocol = property(__get_ip_protocol, __set_ip_protocol)
    """:class:`.IPProtocol`. IP protocol used in the transmission."""

    data = property(__get_data, __set_data)
    """Bytearray. Data of the packet."""


class TXIPv4Packet(XBeeAPIPacket):
    """
    This class represents an TX (Transmit) IPv4 packet. Packet is built
    using the parameters of the constructor or providing a valid byte array.

    .. seealso::
       | :class:`.RXIPv4Packet`
       | :class:`.XBeeAPIPacket`
    """

    OPTIONS_CLOSE_SOCKET = 2
    """This option will close the socket after the transmission."""

    OPTIONS_LEAVE_SOCKET_OPEN = 0
    """This option will leave socket open after the transmission."""

    __MIN_PACKET_LENGTH = 16

    def __init__(self, frame_id, dest_address, dest_port, source_port, ip_protocol, transmit_options, data=None):
        """
        Class constructor. Instantiates a new :class:`.TXIPv4Packet` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID. Must be between 0 and 255.
            dest_address (:class:`.IPv4Address`): IPv4 address of the destination device.
            dest_port (Integer): destination port number.
            source_port (Integer): source port number.
            ip_protocol (:class:`.IPProtocol`): IP protocol used for transmitted data.
            transmit_options (Integer): the transmit options of the packet.
            data (Bytearray, optional): data that is sent to the destination device. Optional.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.
            ValueError: if ``dest_port`` is less than 0 or greater than 65535.
            ValueError: if ``source_port`` is less than 0 or greater than 65535.

        .. seealso::
           | :class:`.IPProtocol`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255")
        if dest_port < 0 or dest_port > 65535:
            raise ValueError("Destination port must be between 0 and 65535")
        if source_port < 0 or source_port > 65535:
            raise ValueError("Source port must be between 0 and 65535")

        super().__init__(ApiFrameType.TX_IPV4)
        self._frame_id = frame_id
        self.__dest_address = dest_address
        self.__dest_port = dest_port
        self.__source_port = source_port
        self.__ip_protocol = ip_protocol
        self.__transmit_options = transmit_options
        self.__data = data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            TXIPv4Packet.

        Raises:
            InvalidPacketException: if the bytearray length is less than 16. (start delim + length (2 bytes) + frame
                type + frame id + dest address (4 bytes) + dest port (2 bytes) + source port (2 bytes) + network
                protocol + transmit options + checksum = 16 bytes)
            InvalidPacketException: if the length field of ``raw`` is different than its real length. (length field:
                bytes 2 and 3)
            InvalidPacketException: if the first byte of ``raw`` is not the header byte. See :class:`.SPECIAL_BYTE`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is not :attr:`ApiFrameType.TX_IPV4`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=TXIPv4Packet.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.TX_IPV4.code:
            raise InvalidPacketException("This packet is not an TXIPv4Packet.")

        return TXIPv4Packet(raw[4], IPv4Address(bytes(raw[5:9])), utils.bytes_to_int(raw[9:11]),
                            utils.bytes_to_int(raw[11:13]), IPProtocol.get(raw[13]),
                            raw[14], raw[15:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def __get_dest_address(self):
        """
        Returns the IPv4 address of the destination device.

        Returns:
            :class:`ipaddress.IPv4Address`: the IPv4 address of the destination device.
        """
        return self.__dest_address

    def __set_dest_address(self, dest_address):
        """
        Sets the IPv4 destination address.

        Args:
            dest_address (:class:`ipaddress.IPv4Address`): The new IPv4 destination address.
        """
        if dest_address is not None:
            self.__dest_address = dest_address

    def __get_dest_port(self):
        """
        Returns the destination port.

        Returns:
            Integer: the destination port.
        """
        return self.__dest_port

    def __set_dest_port(self, dest_port):
        """
        Sets the destination port.

        Args:
            dest_port (Integer): the new destination port.

        Raises:
            ValueError: if ``dest_port`` is less than 0 or greater than 65535.
        """
        if dest_port < 0 or dest_port > 65535:
            raise ValueError("Destination port must be between 0 and 65535")
        self.__dest_port = dest_port

    def __get_source_port(self):
        """
        Returns the source port.

        Returns:
            Integer: the source port.
        """
        return self.__source_port

    def __set_source_port(self, source_port):
        """
        Sets the source port.

        Args:
            source_port (Integer): the new source port.

        Raises:
            ValueError: if ``source_port`` is less than 0 or greater than 65535.
        """
        if source_port < 0 or source_port > 65535:
            raise ValueError("Source port must be between 0 and 65535")

        self.__source_port = source_port

    def __get_ip_protocol(self):
        """
        Returns the IP protocol used for transmitted data.

        Returns:
            :class:`.IPProtocol`: the IP protocol used for transmitted data.
        """
        return self.__ip_protocol

    def __set_ip_protocol(self, ip_protocol):
        """
        Sets the network protocol used for transmitted data.

        Args:
            ip_protocol (:class:`.IPProtocol`): the new IP protocol.
        """
        self.__ip_protocol = ip_protocol

    def __get_transmit_options(self):
        """
        Returns the transmit options of the packet.

        Returns:
            Integer: the transmit options of the packet.
        """
        return self.__transmit_options

    def __set_transmit_options(self, transmit_options):
        """
        Sets the transmit options bitfield of the packet.

        Args:
            transmit_options (Integer): the new transmit options. Can
                be :attr:`OPTIONS_CLOSE_SOCKET` or :attr:`OPTIONS_LEAVE_SOCKET_OPEN`.
        """
        self.__transmit_options = transmit_options

    def __get_data(self):
        """
        Returns the data of the packet.

        Returns:
            Bytearray: the data of the packet.
        """
        return self.__data if self.__data is None else self.__data.copy()

    def __set_data(self, data):
        """
        Sets the data of the packet.

        Args:
            data (Bytearray): the new data of the packet.
        """
        self.__data = None if data is None else data.copy()

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data`
        """
        ret = bytearray(self.__dest_address.packed)
        ret += utils.int_to_bytes(self.__dest_port, num_bytes=2)
        ret += utils.int_to_bytes(self.__source_port, num_bytes=2)
        ret += utils.int_to_bytes(self.__ip_protocol.code)
        ret += utils.int_to_bytes(self.__transmit_options)
        if self.__data is not None:
            ret += self.__data
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data_dict`
        """
        return {DictKeys.DEST_IPV4_ADDR: "%s (%s)" % (self.__dest_address.packed, self.__dest_address.exploded),
                DictKeys.DEST_PORT:      self.dest_port,
                DictKeys.SRC_PORT:       self.source_port,
                DictKeys.IP_PROTOCOL:    "%s (%s)" % (self.__ip_protocol.code, self.__ip_protocol.description),
                DictKeys.OPTIONS:        self.__transmit_options,
                DictKeys.RF_DATA:        bytearray(self.__data)}

    dest_address = property(__get_dest_address, __set_dest_address)
    """:class:`ipaddress.IPv4Address`. IPv4 address of the destination device."""

    dest_port = property(__get_dest_port, __set_dest_port)
    """Integer. Destination port."""

    source_port = property(__get_source_port, __set_source_port)
    """Integer. Source port."""

    ip_protocol = property(__get_ip_protocol, __set_ip_protocol)
    """:class:`.IPProtocol`. IP protocol."""

    transmit_options = property(__get_transmit_options, __set_transmit_options)
    """Integer. Transmit options."""

    data = property(__get_data, __set_data)
    """Bytearray. Data of the packet."""



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.models.hw import HardwareVersion
from digi.xbee.util import utils


@unique
class XBeeProtocol(Enum):
    """
    Enumerates the available XBee protocols. The XBee protocol is determined
    by the combination of hardware and firmware of an XBee device.

    | Inherited properties:
    |     **name** (String): the name (id) of this XBeeProtocol.
    |     **value** (String): the value of this XBeeProtocol.
    """

    ZIGBEE = (0, "ZigBee")
    RAW_802_15_4 = (1, "802.15.4")
    XBEE_WIFI = (2, "Wi-Fi")
    DIGI_MESH = (3, "DigiMesh")
    XCITE = (4, "XCite")
    XTEND = (5, "XTend (Legacy)")
    XTEND_DM = (6, "XTend (DigiMesh)")
    SMART_ENERGY = (7, "Smart Energy")
    DIGI_POINT = (8, "Point-to-multipoint")
    ZNET = (9, "ZNet 2.5")
    XC = (10, "XSC")
    XLR = (11, "XLR")
    XLR_DM = (12, "XLR")
    SX = (13, "XBee SX")
    XLR_MODULE = (14, "XLR Module")
    CELLULAR = (15, "Cellular")
    CELLULAR_NBIOT = (16, "Cellular NB-IoT")
    UNKNOWN = (99, "Unknown")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the XBeeProtocol element.

        Returns:
            Integer: the code of the XBeeProtocol element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the XBeeProtocol element.

        Returns:
            String: the description of the XBeeProtocol element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the XBeeProtocol for the given code.

        Args:
            code (Integer): code of the XBeeProtocol to get.

        Returns:
            XBeeProtocol: XBeeProtocol for the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return XBeeProtocol.UNKNOWN

    @staticmethod
    def determine_protocol(hardware_version, firmware_version):
        """
        Determines the XBee protocol based on the given hardware and firmware
        versions.

        Args:
            hardware_version (Integer): hardware version of the protocol to determine.
            firmware_version (String): firmware version of the protocol to determine.

        Returns:
            The XBee protocol corresponding to the given hardware and firmware versions.
        """
        firmware_version = "".join(["%02X" % i for i in firmware_version])

        if hardware_version is None or firmware_version is None or hardware_version < 0x09 or \
                HardwareVersion.get(hardware_version) is None:
            return XBeeProtocol.UNKNOWN

        elif hardware_version in [HardwareVersion.XC09_009.code,
                                  HardwareVersion.XC09_038.code]:
            return XBeeProtocol.XCITE

        elif hardware_version in [HardwareVersion.XT09_XXX.code,
                                  HardwareVersion.XT09B_XXX.code]:
            if ((len(firmware_version) == 4 and firmware_version.startswith("8")) or
                    (len(firmware_version) == 5 and firmware_version[1] == '8')):
                return XBeeProtocol.XTEND_DM
            return XBeeProtocol.XTEND

        elif hardware_version in [HardwareVersion.XB24_AXX_XX.code,
                                  HardwareVersion.XBP24_AXX_XX.code]:
            if len(firmware_version) == 4 and firmware_version.startswith("8"):
                    return XBeeProtocol.DIGI_MESH
            return XBeeProtocol.RAW_802_15_4

        elif hardware_version in [HardwareVersion.XB24_BXIX_XXX.code,
                                  HardwareVersion.XBP24_BXIX_XXX.code]:
            if ((len(firmware_version) == 4 and firmware_version.startswith("1") and firmware_version.endswith("20"))
                    or (len(firmware_version) == 4 and firmware_version.startswith("2"))):
                return XBeeProtocol.ZIGBEE
            elif len(firmware_version) == 4 and firmware_version.startswith("3"):
                return XBeeProtocol.SMART_ENERGY
            return XBeeProtocol.ZNET

        elif hardware_version == HardwareVersion.XBP09_DXIX_XXX.code:
            if ((len(firmware_version) == 4 and firmware_version.startswith("8") or
                (len(firmware_version) == 4 and firmware_version[1] == '8')) or
                    (len(firmware_version) == 5 and firmware_version[1] == '8')):
                return XBeeProtocol.DIGI_MESH
            return XBeeProtocol.DIGI_POINT

        elif hardware_version == HardwareVersion.XBP09_XCXX_XXX.code:
            return XBeeProtocol.XC

        elif hardware_version == HardwareVersion.XBP08_DXXX_XXX.code:
            return XBeeProtocol.DIGI_POINT

        elif hardware_version == HardwareVersion.XBP24B.code:
            if len(firmware_version) == 4 and firmware_version.startswith("3"):
                return XBeeProtocol.SMART_ENERGY
            return XBeeProtocol.ZIGBEE

        elif hardware_version in [HardwareVersion.XB24_WF.code,
                                  HardwareVersion.WIFI_ATHEROS.code,
                                  HardwareVersion.SMT_WIFI_ATHEROS.code]:
            return XBeeProtocol.XBEE_WIFI

        elif hardware_version in [HardwareVersion.XBP24C.code,
                                  HardwareVersion.XB24C.code]:
            if (len(firmware_version) == 4 and (firmware_version.startswith("5")) or
                    (firmware_version.startswith("6"))):
                return XBeeProtocol.SMART_ENERGY
            elif firmware_version.startswith("2"):
                return XBeeProtocol.RAW_802_15_4
            elif firmware_version.startswith("9"):
                return XBeeProtocol.DIGI_MESH
            return XBeeProtocol.ZIGBEE

        elif hardware_version in [HardwareVersion.XSC_GEN3.code,
                                  HardwareVersion.SRD_868_GEN3.code]:
            if len(firmware_version) == 4 and firmware_version.startswith("8"):
                return XBeeProtocol.DIGI_MESH
            elif len(firmware_version) == 4 and firmware_version.startswith("1"):
                return XBeeProtocol.DIGI_POINT
            return XBeeProtocol.XC

        elif hardware_version == HardwareVersion.XBEE_CELL_TH.code:
            return XBeeProtocol.UNKNOWN

        elif hardware_version == HardwareVersion.XLR_MODULE.code:
            # This is for the old version of the XLR we have (K60), and it is
            # reporting the firmware of the module (8001), this will change in
            # future (after K64 integration) reporting the hardware and firmware
            # version of the baseboard (see the case HardwareVersion.XLR_BASEBOARD).
            # TODO maybe this should be removed in future, since this case will never be released.
            if firmware_version.startswith("1"):
                return XBeeProtocol.XLR
            else:
                return XBeeProtocol.XLR_MODULE

        elif hardware_version == HardwareVersion.XLR_BASEBOARD.code:
            # XLR devices with K64 will report the baseboard hardware version,
            # and also firmware version (the one we have here is 1002, but this value
            # is not being reported since is an old K60 version, the module fw version
            # is reported instead).

            # TODO [XLR_DM] The next version of the XLR will add DigiMesh support should be added.
            # Probably this XLR_DM and XLR will depend on the firmware version.
            if firmware_version.startswith("1"):
                return XBeeProtocol.XLR
            else:
                return XBeeProtocol.XLR_MODULE

        elif hardware_version == HardwareVersion.XB900HP_NZ.code:
            return XBeeProtocol.DIGI_POINT

        elif hardware_version in [HardwareVersion.XBP24C_TH_DIP.code,
                                  HardwareVersion.XB24C_TH_DIP.code,
                                  HardwareVersion.XBP24C_S2C_SMT.code]:
            if (len(firmware_version) == 4 and
                    (firmware_version.startswith("5") or firmware_version.startswith("6"))):
                return XBeeProtocol.SMART_ENERGY
            elif firmware_version.startswith("2"):
                return XBeeProtocol.RAW_802_15_4
            elif firmware_version.startswith("9"):
                return XBeeProtocol.DIGI_MESH
            return XBeeProtocol.ZIGBEE

        elif hardware_version in [HardwareVersion.SX_PRO.code,
                                  HardwareVersion.SX.code,
                                  HardwareVersion.XTR.code]:
            if firmware_version.startswith("2"):
                return XBeeProtocol.XTEND
            elif firmware_version.startswith("8"):
                return XBeeProtocol.XTEND_DM
            return XBeeProtocol.DIGI_MESH

        elif hardware_version in [HardwareVersion.S2D_SMT_PRO.code,
                                  HardwareVersion.S2D_SMT_REG.code,
                                  HardwareVersion.S2D_TH_PRO.code,
                                  HardwareVersion.S2D_TH_REG.code]:
            return XBeeProtocol.ZIGBEE

        elif hardware_version in [HardwareVersion.CELLULAR_CAT1_LTE_VERIZON.code,
                                  HardwareVersion.CELLULAR_3G.code,
                                  HardwareVersion.CELLULAR_LTE_ATT.code,
                                  HardwareVersion.CELLULAR_LTE_VERIZON.code,
                                  HardwareVersion.CELLULAR_3_CAT1_LTE_ATT.code,
                                  HardwareVersion.CELLULAR_3_LTE_M_VERIZON.code,
                                  HardwareVersion.CELLULAR_3_LTE_M_ATT.code]:
            return XBeeProtocol.CELLULAR

        elif hardware_version == HardwareVersion.CELLULAR_NBIOT_EUROPE.code:
            return XBeeProtocol.CELLULAR_NBIOT

        elif hardware_version in [HardwareVersion.XBEE3.code,
                                  HardwareVersion.XBEE3_SMT.code,
                                  HardwareVersion.XBEE3_TH.code]:
            if firmware_version.startswith("2"):
                return XBeeProtocol.RAW_802_15_4
            elif firmware_version.startswith("3"):
                return XBeeProtocol.DIGI_MESH
            else:
                return XBeeProtocol.ZIGBEE

        elif hardware_version == HardwareVersion.XB8X.code:
            return XBeeProtocol.DIGI_MESH

        return XBeeProtocol.ZIGBEE

    code = property(__get_code)
    """Integer. XBee protocol code."""

    description = property(__get_description)
    """String. XBee protocol description."""


XBeeProtocol.lookupTable = {x.code: x for x in XBeeProtocol}
XBeeProtocol.__doc__ += utils.doc_enum(XBeeProtocol)


@unique
class IPProtocol(Enum):
    """
    Enumerates the available network protocols.

    | Inherited properties:
    |     **name** (String): the name (id) of this IPProtocol.
    |     **value** (String): the value of this IPProtocol.
    """

    UDP = (0, "UDP")
    TCP = (1, "TCP")
    TCP_SSL = (4, "TCP SSL")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the IP protocol.

        Returns:
            Integer: code of the IP protocol.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the IP protocol.

        Returns:
            String: description of the IP protocol.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the IPProtocol for the given code.

        Args:
            code (Integer): code associated to the IP protocol.

        Returns:
            :class:`.IPProtocol`: IP protocol for the given code or ``None`` if there
                is not any ``IPProtocol`` with the given code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer: IP protocol code."""

    description = property(__get_description)
    """String: IP protocol description."""


IPProtocol.lookupTable = {x.code: x for x in IPProtocol}
IPProtocol.__doc__ += utils.doc_enum(IPProtocol)



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from ipaddress import IPv4Address
from digi.xbee.devices import CellularDevice
from digi.xbee.models.protocol import IPProtocol

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600
# TODO: Optionally, replace with the text you want to send to the server.
TEXT = "Hello XBee!"

ECHO_SERVER = "52.43.121.77"
ECHO_SERVER_PORT = 11001
PROTOCOL = IPProtocol.TCP


def main():
    print(" +---------------------------------------------------+")
    print(" | XBee Python Library Connect to Echo Server Sample |")
    print(" +---------------------------------------------------+\n")

    device = CellularDevice(PORT, BAUD_RATE)

    try:
        device.open()

        if not device.is_connected():
            print(">> Error: the device is not connected to the network")
            return

        print("Sending text to %s:%d >> %s..." % (ECHO_SERVER, ECHO_SERVER_PORT, TEXT))

        device.send_ip_data(IPv4Address(ECHO_SERVER), ECHO_SERVER_PORT, PROTOCOL, TEXT)

        print("Success")

        # Read the echoed data.
        ip_message = device.read_ip_data()
        if ip_message is None:
            print("Echo response was not received from the server.")
            return

        print("Echo response received from %s:%d >> '%s'" % (ip_message.ip_addr, ip_message.source_port,
                                                             ip_message.data.decode("utf8")))

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.models.status import ATCommandStatus
from enum import Enum, unique
from digi.xbee.util import utils


@unique
class ATStringCommand(Enum):
    """
    This class represents basic AT commands.

    | Inherited properties:
    |     **name** (String): name (ID) of this ATStringCommand.
    |     **value** (String): value of this ATStringCommand.
    """

    NI = "NI"
    KY = "KY"
    NK = "NK"
    ZU = "ZU"
    ZV = "ZV"
    CC = "CC"

    def __init__(self, command):
        self.__command = command

    def __get_command(self):
        return self.__command

    command = property(__get_command)
    """String. AT Command alias."""


ATStringCommand.__doc__ += utils.doc_enum(ATStringCommand)


@unique
class SpecialByte(Enum):
    """
    Enumerates all the special bytes of the XBee protocol that must be escaped
    when working on API 2 mode.

    | Inherited properties:
    |     **name** (String): name (ID) of this SpecialByte.
    |     **value** (String): the value of this SpecialByte.
    """

    ESCAPE_BYTE = 0x7D
    HEADER_BYTE = 0x7E
    XON_BYTE = 0x11
    XOFF_BYTE = 0x13

    def __init__(self, code):
        self.__code = code

    def __get_code(self):
        """
        Returns the code of the SpecialByte element.

        Returns:
            Integer: the code of the SpecialByte element.
        """
        return self.__code

    @classmethod
    def get(cls, value):
        """
        Returns the special byte for the given value.

        Args:
            value (Integer): value associated to the special byte.

        Returns:
            SpecialByte: SpecialByte with the given value.
        """
        return SpecialByte.lookupTable[value]

    @staticmethod
    def escape(value):
        """
        Escapes the byte by performing a XOR operation with 0x20 value.

        Args:
            value (Integer): value to escape.

        Returns:
            Integer: value ^ 0x20 (escaped).
        """
        return value ^ 0x20

    @staticmethod
    def is_special_byte(value):
        """
        Checks whether the given byte is special or not.

        Args:
            value (Integer): byte to check.

        Returns:
            Boolean: ``True`` if value is a special byte, ``False`` in other case.
        """
        return True if value in [i.value for i in SpecialByte] else False

    code = property(__get_code)
    """Integer. The special byte code."""


SpecialByte.lookupTable = {x.code: x for x in SpecialByte}
SpecialByte.__doc__ += utils.doc_enum(SpecialByte)


class ATCommand(object):
    """
    This class represents an AT command used to read or set different properties
    of the XBee device.

    AT commands can be sent directly to the connected device or to remote
    devices and may have parameters.

    After executing an AT Command, an AT Response is received from the device.
    """

    def __init__(self, command, parameter=None):
        """
        Class constructor. Instantiates a new :class:`.ATCommand` object with the provided parameters.

        Args:
            command (String): AT Command, must have length 2.
            parameter (String or Bytearray, optional): The AT parameter value. Defaults to ``None``. Optional.

        Raises:
            ValueError: if command length is not 2.
        """
        if len(command) != 2:
            raise ValueError("Command length must be 2.")

        self.__command = command
        self.__set_parameter(parameter)

    def __str__(self):
        """
        Returns a string representation of this ATCommand.

        Returns:
            String: representation of this ATCommand.
        """
        return "Command: " + self.__command + "\n" + "Parameter: " + str(self.__parameter)

    def __len__(self):
        """
        Returns the length of this ATCommand.

        Returns:
            Integer: length of command + length of parameter.
        """
        if self.__parameter:
            return len(self.__command) + len(self.__parameter)
        else:
            return len(self.__command)

    def __get_command(self):
        """
        Returns the AT command.

        Returns:
            ATCommand: the AT command.
        """
        return self.__command

    def __get_parameter(self):
        """
        Returns the AT command parameter.

        Returns:
            Bytearray: the AT command parameter. ``None`` if there is no parameter.
        """
        return self.__parameter

    def get_parameter_string(self):
        """
        Returns this ATCommand parameter as a String.

        Returns:
            String: this ATCommand parameter. ``None`` if there is no parameter.
        """
        return self.__parameter.decode() if self.__parameter else None

    def __set_parameter(self, parameter):
        """
        Sets the AT command parameter.

        Args:
            parameter (Bytearray): the parameter to be set.
        """
        if isinstance(parameter, str):
            self.__parameter = bytearray(parameter, 'utf8')
        else:
            self.__parameter = parameter

    command = property(__get_command)
    """String. The AT command"""

    parameter = property(__get_parameter, __set_parameter)
    """Bytearray. The AT command parameter"""


class ATCommandResponse(object):
    """
    This class represents the response of an AT Command sent by the connected
    XBee device or by a remote device after executing an AT Command.
    """

    def __init__(self, command, response=None, status=ATCommandStatus.OK):
        """
        Class constructor.

        Args:
            command (ATCommand): The AT command that generated the response.
            response (bytearray, optional): The command response. Default to ``None``.
            status (ATCommandStatus, optional): The AT command status. Default to ATCommandStatus.OK
        """
        self.__atCommand = command
        self.__response = response
        self.__comm_status = status

    def __get_command(self):
        """
        Returns the AT command.

        Returns:
            ATCommand: the AT command.
        """
        return self.__atCommand

    def __get_response(self):
        """
        Returns the AT command response.

        Returns:
            Bytearray: the AT command response.
        """
        return self.__response

    def __get_status(self):
        """
        Returns the AT command response status.

        Returns:
            ATCommandStatus: The AT command response status.
        """
        return self.__comm_status

    command = property(__get_command)
    """String. The AT command."""

    response = property(__get_response)
    """Bytearray. The AT command response data."""

    status = property(__get_status)
    """ATCommandStatus. The AT command response status."""



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.packets.base import XBeeAPIPacket, DictKeys
from digi.xbee.exception import InvalidOperatingModeException, InvalidPacketException
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.models.mode import OperatingMode
from digi.xbee.models.options import TransmitOptions
from digi.xbee.util import utils
import re


PATTERN_PHONE_NUMBER = "^\+?\d+$"
"""Pattern used to validate the phone number parameter of SMS packets."""


class RXSMSPacket(XBeeAPIPacket):
    """
    This class represents an RX (Receive) SMS packet. Packet is built 
    using the parameters of the constructor or providing a valid byte array.
    
    .. seealso::
       | :class:`.TXSMSPacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 25

    def __init__(self, phone_number, data):
        """
        Class constructor. Instantiates a new :class:`.RXSMSPacket` object withe the provided parameters.
        
        Args:
            phone_number (String): phone number of the device that sent the SMS.
            data (String): packet data (text of the SMS).
            
        Raises:
            ValueError: if length of ``phone_number`` is greater than 20.
            ValueError: if ``phone_number`` is not a valid phone number.
        """
        if len(phone_number) > 20:
            raise ValueError("Phone number length cannot be greater than 20 bytes")
        if not re.match(PATTERN_PHONE_NUMBER, phone_number):
            raise ValueError("Phone number invalid, only numbers and '+' prefix allowed.")
        super().__init__(ApiFrameType.RX_SMS)

        self.__phone_number = bytearray(20)
        self.__phone_number[0:len(phone_number)] = phone_number.encode("utf8")
        self.__data = data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.
        
        Returns:
            :class:`.RXSMSPacket`
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 25. (start delim + length (2 bytes) +
                frame type + phone number (20 bytes) + checksum = 25 bytes)
            InvalidPacketException: if the length field of ``raw`` is different than its real length. (length field:
                bytes 2 and 3)
            InvalidPacketException: if the first byte of ``raw`` is not the header byte. See :class:`.SPECIAL_BYTE`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :py:attr:`.ApiFrameType.RX_SMS`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.
            
        .. seealso::
           | :meth:`.XBeePacket.create_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")
        
        XBeeAPIPacket._check_api_packet(raw, min_length=RXSMSPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.RX_SMS.code:
            raise InvalidPacketException("This packet is not an RXSMSPacket")

        return RXSMSPacket(raw[4:23].decode("utf8").replace("\0", ""), raw[24:-1].decode("utf8"))

    def needs_id(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def get_phone_number_byte_array(self):
        """
        Returns the phone number byte array.

        Returns:
            Bytearray: phone number of the device that sent the SMS.
        """
        return self.__phone_number

    def __get_phone_number(self):
        """
        Returns the phone number of the device that sent the SMS.

        Returns:
            String: phone number of the device that sent the SMS.
        """
        return self.__phone_number.decode("utf8").replace("\0", "")

    def __set_phone_number(self, phone_number):
        """
        Sets the phone number of the device that sent the SMS.

        Args:
            phone_number (String): the new phone number.

        Raises:
            ValueError: if length of ``phone_number`` is greater than 20.
            ValueError: if ``phone_number`` is not a valid phone number.
        """
        if len(phone_number) > 20:
            raise ValueError("Phone number length cannot be greater than 20 bytes")
        if not re.match(PATTERN_PHONE_NUMBER, phone_number):
            raise ValueError("Phone number invalid, only numbers and '+' prefix allowed.")

        self.__phone_number = bytearray(20)
        self.__phone_number[0:len(phone_number)] = phone_number.encode("utf8")

    def __get_data(self):
        """
        Returns the data of the packet (SMS text).

        Returns:
            String: the data of the packet.
        """
        return self.__data

    def __set_data(self, data):
        """
        Sets the data of the packet.

        Args:
            data (String): the new data of the packet.
        """
        self.__data = data

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data`
        """
        ret = bytearray()
        ret += self.__phone_number
        if self.__data is not None:
            ret += self.__data.encode("utf8")
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data_dict`
        """
        return {DictKeys.PHONE_NUMBER: self.__phone_number,
                DictKeys.RF_DATA:      self.__data}

    phone_number = property(__get_phone_number, __set_phone_number)
    """String. Phone number that sent the SMS."""

    data = property(__get_data, __set_data)
    """String. Data of the SMS."""


class TXSMSPacket(XBeeAPIPacket):
    """
    This class represents a TX (Transmit) SMS packet. Packet is built 
    using the parameters of the constructor or providing a valid byte array.
    
    .. seealso::
       | :class:`.RXSMSPacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 27

    def __init__(self, frame_id, phone_number, data):
        """
        Class constructor. Instantiates a new :class:`.TXSMSPacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID. Must be between 0 and 255.
            phone_number (String): the phone number.
            data (String): this packet's data.

        Raises:
            ValueError: if ``frame_id`` is not between 0 and 255.
            ValueError: if length of ``phone_number`` is greater than 20.
            ValueError: if ``phone_number`` is not a valid phone number.

        .. seealso::
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255")
        if len(phone_number) > 20:
            raise ValueError("Phone number length cannot be greater than 20 bytes")
        if not re.match(PATTERN_PHONE_NUMBER, phone_number):
            raise ValueError("Phone number invalid, only numbers and '+' prefix allowed.")
        super().__init__(ApiFrameType.TX_SMS)
        
        self._frame_id = frame_id
        self.__transmit_options = TransmitOptions.NONE.value
        self.__phone_number = bytearray(20)
        self.__phone_number[0:len(phone_number)] = phone_number.encode("utf8")
        self.__data = data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.TXSMSPacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 27. (start delim, length (2 bytes), frame type,
                frame id, transmit options, phone number (20 bytes), checksum)
            InvalidPacketException: if the length field of ``raw`` is different than its real length. (length field:
                bytes 2 and 3)
            InvalidPacketException: if the first byte of ``raw`` is not the header byte. See :class:`.SPECIAL_BYTE`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :py:attr:`.ApiFrameType.TX_SMS`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")
        
        XBeeAPIPacket._check_api_packet(raw, min_length=TXSMSPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.TX_SMS.code:
            raise InvalidPacketException("This packet is not a TXSMSPacket")

        return TXSMSPacket(raw[4], raw[6:25].decode("utf8").replace("\0", ""), raw[26:-1].decode("utf8"))

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def get_phone_number_byte_array(self):
        """
        Returns the phone number byte array.

        Returns:
            Bytearray: phone number of the device that sent the SMS.
        """
        return self.__phone_number

    def __get_phone_number(self):
        """
        Returns the phone number of the transmitter device.

        Returns:
            String: the phone number of the transmitter device.
        """
        return self.__phone_number.decode("utf8").replace("\0", "")

    def __set_phone_number(self, phone_number):
        """
        Sets the phone number of the transmitter device.

        Args:
            phone_number (String): the new phone number.

        Raises:
            ValueError: if length of ``phone_number`` is greater than 20.
            ValueError: if ``phone_number`` is not a valid phone number.
        """
        if len(phone_number) > 20:
            raise ValueError("Phone number length cannot be greater than 20 bytes")
        if not re.match(PATTERN_PHONE_NUMBER, phone_number):
            raise ValueError("Phone number invalid, only numbers and '+' prefix allowed.")

        self.__phone_number = bytearray(20)
        self.__phone_number[0:len(phone_number)] = phone_number.encode("utf8")

    def __get_data(self):
        """
        Returns the data of the packet (SMS text).

        Returns:
            Bytearray: packet's data.
        """
        return self.__data

    def __set_data(self, data):
        """
        Sets the data of the packet.

        Args:
            data (Bytearray): the new data of the packet.
        """
        self.__data = data

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data`
        """
        ret = utils.int_to_bytes(self.__transmit_options, num_bytes=1)
        ret += self.__phone_number
        if self.__data is not None:
            ret += self.__data.encode("utf8")
        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_API_packet_spec_data_dict`
        """
        return {DictKeys.OPTIONS:      self.__transmit_options,
                DictKeys.PHONE_NUMBER: self.__phone_number,
                DictKeys.RF_DATA:      self.__data}

    phone_number = property(__get_phone_number, __set_phone_number)
    """String. Phone number that sent the SMS."""

    data = property(__get_data, __set_data)
    """String. Data of the SMS."""



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.packets.base import XBeeAPIPacket, DictKeys
from digi.xbee.util import utils
from digi.xbee.exception import InvalidOperatingModeException, InvalidPacketException
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.models.mode import OperatingMode
from digi.xbee.models.status import DeviceCloudStatus, FrameError
from digi.xbee.models.options import SendDataRequestOptions


class DeviceRequestPacket(XBeeAPIPacket):
    """
    This class represents a device request packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is sent out the serial port when the XBee module receives
    a valid device request from Device Cloud.

    .. seealso::
       | :class:`.DeviceResponsePacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 5

    def __init__(self, request_id, target=None, request_data=None):
        """
        Class constructor. Instantiates a new :class:`.DeviceRequestPacket` object with the provided parameters.

        Args:
            request_id (Integer): number that identifies the device request. (0 has no special meaning)
            target (String): device request target.
            request_data (Bytearray, optional): data of the request. Optional.

        Raises:
            ValueError: if ``request_id`` is less than 0 or greater than 255.
            ValueError: if length of ``target`` is greater than 255.

        .. seealso::
           | :class:`.XBeeAPIPacket`
        """
        if request_id < 0 or request_id > 255:
            raise ValueError("Device request ID must be between 0 and 255.")

        if target is not None and len(target) > 255:
            raise ValueError("Target length cannot exceed 255 bytes.")

        super().__init__(ApiFrameType.DEVICE_REQUEST)
        self.__request_id = request_id
        self.__transport = 0x00  # Reserved.
        self.__flags = 0x00  # Reserved.
        self.__target = target
        self.__request_data = request_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.DeviceRequestPacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 9. (start delim. + length (2 bytes) + frame
                type + request id + transport + flags + target length + checksum = 9 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.DEVICE_REQUEST`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=DeviceRequestPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.DEVICE_REQUEST.code:
            raise InvalidPacketException("This packet is not a device request packet.")

        target_length = raw[7]
        return DeviceRequestPacket(raw[4], raw[8:8 + target_length].decode("utf8"), raw[8 + target_length:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = utils.int_to_bytes(self.__request_id, num_bytes=1)
        ret += utils.int_to_bytes(self.__transport, num_bytes=1)
        ret += utils.int_to_bytes(self.__flags, num_bytes=1)
        if self.__target is not None:
            ret += utils.int_to_bytes(len(self.__target), num_bytes=1)
            ret += bytearray(self.__target, "utf8")
        else:
            ret += utils.int_to_bytes(0x00, num_bytes=1)
        if self.__request_data is not None:
            ret += self.__request_data

        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        See:
            :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.REQUEST_ID: self.__request_id,
                DictKeys.TRANSPORT:  self.__transport,
                DictKeys.FLAGS:      self.__flags,
                DictKeys.TARGET:     self.__target,
                DictKeys.RF_DATA:    list(self.__request_data) if self.__request_data is not None else None}

    def __get_request_id(self):
        """
        Returns the request ID of the packet.

        Returns:
            Integer: the request ID of the packet.
        """
        return self.__request_id

    def __set_request_id(self, request_id):
        """
        Sets the request ID of the packet.

        Args:
            request_id (Integer): the new request ID of the packet. Must be between 0 and 255.

        Raises:
            ValueError: if ``request_id`` is less than 0 or greater than 255.
        """
        if request_id < 0 or request_id > 255:
            raise ValueError("Device request ID must be between 0 and 255.")
        self.__request_id = request_id

    def __get_transport(self):
        """
        Returns the transport of the packet.

        Returns:
            Integer: the transport of the packet.
        """
        return self.__transport

    def __get_flags(self):
        """
        Returns the flags of the packet.

        Returns:
            Integer: the flags of the packet.
        """
        return self.__flags

    def __get_target(self):
        """
        Returns the device request target of the packet.

        Returns:
            String: the device request target of the packet.
        """
        return self.__target

    def __set_target(self, target):
        """
        Sets the device request target of the packet.

        Args:
            target (String): the new device request target of the packet.

        Raises:
            ValueError: if ``target`` length is greater than 255.
        """
        if target is not None and len(target) > 255:
            raise ValueError("Target length cannot exceed 255 bytes.")
        self.__target = target

    def __get_request_data(self):
        """
        Returns the data of the device request.

        Returns:
            Bytearray: the data of the device request.
        """
        if self.__request_data is None:
            return None
        return self.__request_data.copy()

    def __set_request_data(self, request_data):
        """
        Sets the data of the device request.

        Args:
            request_data (Bytearray): the new data of the device request.
        """
        if request_data is None:
            self.__request_data = None
        else:
            self.__request_data = request_data.copy()

    request_id = property(__get_request_id, __set_request_id)
    """Integer. Request ID of the packet."""

    transport = property(__get_transport)
    """Integer. Transport (reserved)."""

    flags = property(__get_flags)
    """Integer. Flags (reserved)."""

    target = property(__get_target, __set_target)
    """String. Request target of the packet."""

    request_data = property(__get_request_data, __set_request_data)
    """Bytearray. Data of the device request."""


class DeviceResponsePacket(XBeeAPIPacket):
    """
    This class represents a device response packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is sent to the serial port by the host in response to the
    :class:`.DeviceRequestPacket`. It should be sent within five seconds to avoid
    a timeout error.

    .. seealso::
       | :class:`.DeviceRequestPacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 8

    def __init__(self, frame_id, request_id, response_data=None):
        """
        Class constructor. Instantiates a new :class:`.DeviceResponsePacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            request_id (Integer): device Request ID. This number should match the device request ID in the
                device request. Otherwise, an error will occur. (0 has no special meaning)
            response_data (Bytearray, optional): data of the response. Optional.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.
            ValueError: if ``request_id`` is less than 0 or greater than 255.

        .. seealso::
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")
        if request_id < 0 or request_id > 255:
            raise ValueError("Device request ID must be between 0 and 255.")

        super().__init__(ApiFrameType.DEVICE_RESPONSE)
        self._frame_id = frame_id
        self.__request_id = request_id
        self.__response_data = response_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.DeviceResponsePacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 8. (start delim. + length (2 bytes) + frame
                type + frame id + request id + reserved + checksum = 8 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.DEVICE_RESPONSE`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=DeviceResponsePacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.DEVICE_RESPONSE.code:
            raise InvalidPacketException("This packet is not a device response packet.")

        return DeviceResponsePacket(raw[4], raw[5], raw[7:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        ret = utils.int_to_bytes(self.__request_id, num_bytes=1)
        ret += utils.int_to_bytes(0x00, num_bytes=1)
        if self.__response_data is not None:
            ret += self.__response_data

        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.REQUEST_ID: self.__request_id,
                DictKeys.RESERVED:   0x00,
                DictKeys.RF_DATA:    list(self.__response_data) if self.__response_data is not None else None}

    def __get_request_id(self):
        """
        Returns the request ID of the packet.

        Returns:
            Integer: the request ID of the packet.
        """
        return self.__request_id

    def __set_request_id(self, request_id):
        """
        Sets the request ID of the packet.

        Args:
            request_id (Integer): the new request ID of the packet. Must be between 0 and 255.

        Raises:
            ValueError: if ``request_id`` is less than 0 or greater than 255.
        """
        if request_id < 0 or request_id > 255:
            raise ValueError("Device request ID must be between 0 and 255.")
        self.__request_id = request_id

    def __get_response_data(self):
        """
        Returns the data of the device response.

        Returns:
            Bytearray: the data of the device response.
        """
        if self.__response_data is None:
            return None
        return self.__response_data.copy()

    def __set_response_data(self, response_data):
        """
        Sets the data of the device response.

        Args:
            response_data (Bytearray): the new data of the device response.
        """
        if response_data is None:
            self.__response_data = None
        else:
            self.__response_data = response_data.copy()

    request_id = property(__get_request_id, __set_request_id)
    """Integer. Request ID of the packet."""

    request_data = property(__get_response_data, __set_response_data)
    """Bytearray. Data of the device response."""


class DeviceResponseStatusPacket(XBeeAPIPacket):
    """
    This class represents a device response status packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is sent to the serial port after the serial port sends a
    :class:`.DeviceResponsePacket`.

    .. seealso::
       | :class:`.DeviceResponsePacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 7

    def __init__(self, frame_id, status):
        """
        Class constructor. Instantiates a new :class:`.DeviceResponseStatusPacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            status (:class:`.DeviceCloudStatus`): device response status.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.

        .. seealso::
           | :class:`.DeviceCloudStatus`
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.DEVICE_RESPONSE_STATUS)
        self._frame_id = frame_id
        self.__status = status

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.DeviceResponseStatusPacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 7. (start delim. + length (2 bytes) + frame
                type + frame id + device response status + checksum = 7 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.DEVICE_RESPONSE_STATUS`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=DeviceResponseStatusPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.DEVICE_RESPONSE_STATUS.code:
            raise InvalidPacketException("This packet is not a device response status packet.")

        return DeviceResponseStatusPacket(raw[4], DeviceCloudStatus.get(raw[5]))

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return utils.int_to_bytes(self.__status.code, num_bytes=1)

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        See:
            :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.DC_STATUS: self.__status}

    def __get_status(self):
        """
        Returns the status of the device response packet.

        Returns:
            :class:`.DeviceCloudStatus`: the status of the device response packet.

        .. seealso::
           | :class:`.DeviceCloudStatus`
        """
        return self.__status

    def __set_status(self, status):
        """
        Sets the status of the device response packet.

        Args:
            status (:class:`.DeviceCloudStatus`): the new status of the device response packet.

        .. seealso::
           | :class:`.DeviceCloudStatus`
        """
        self.__status = status

    status = property(__get_status, __set_status)
    """:class:`.DeviceCloudStatus`. Status of the device response."""


class FrameErrorPacket(XBeeAPIPacket):
    """
    This class represents a frame error packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is sent to the serial port for any type of frame error.

    .. seealso::
       | :class:`.FrameError`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 6

    def __init__(self, frame_error):
        """
        Class constructor. Instantiates a new :class:`.FrameErrorPacket` object with the provided parameters.

        Args:
            frame_error (:class:`.FrameError`): the frame error.

        .. seealso::
           | :class:`.FrameError`
           | :class:`.XBeeAPIPacket`
        """
        super().__init__(ApiFrameType.FRAME_ERROR)
        self.__frame_error = frame_error

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.FrameErrorPacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 6. (start delim. + length (2 bytes) + frame
                type + frame error + checksum = 6 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.FRAME_ERROR`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=FrameErrorPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.FRAME_ERROR.code:
            raise InvalidPacketException("This packet is not a frame error packet.")

        return FrameErrorPacket(FrameError.get(raw[4]))

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return utils.int_to_bytes(self.__frame_error.code, num_bytes=1)

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.FRAME_ERROR: self.__frame_error}

    def __get_error(self):
        """
        Returns the frame error of the packet.

        Returns:
            :class:`.FrameError`: the frame error of the packet.

        .. seealso::
           | :class:`.FrameError`
        """
        return self.__frame_error

    def __set_error(self, frame_error):
        """
        Sets the frame error of the packet.

        Args:
            frame_error (:class:`.FrameError`): the new frame error of the packet.

        .. seealso::
           | :class:`.FrameError`
        """
        self.__frame_error = frame_error

    error = property(__get_error, __set_error)
    """:class:`.FrameError`. Frame error of the packet."""


class SendDataRequestPacket(XBeeAPIPacket):
    """
    This class represents a send data request packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is used to send a file of the given name and type to
    Device Cloud.

    If the frame ID is non-zero, a :class:`.SendDataResponsePacket` will be
    received.

    .. seealso::
       | :class:`.SendDataResponsePacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 10

    def __init__(self, frame_id, path, content_type, options, file_data=None):
        """
        Class constructor. Instantiates a new :class:`.SendDataRequestPacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            path (String): path of the file to upload to Device Cloud.
            content_type (String): content type of the file to upload.
            options (:class:`.SendDataRequestOptions`): the action when uploading a file.
            file_data (Bytearray, optional): data of the file to upload. Optional.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.

        .. seealso::
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.SEND_DATA_REQUEST)
        self._frame_id = frame_id
        self.__path = path
        self.__content_type = content_type
        self.__transport = 0x00  # Always TCP.
        self.__options = options
        self.__file_data = file_data

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.SendDataRequestPacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 10. (start delim. + length (2 bytes) + frame
                type + frame id + path length + content type length + transport + options + checksum = 10 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.SEND_DATA_REQUEST`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=SendDataRequestPacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.SEND_DATA_REQUEST.code:
            raise InvalidPacketException("This packet is not a send data request packet.")

        path_length = raw[5]
        content_type_length = raw[6 + path_length]
        return SendDataRequestPacket(raw[4],
                                     raw[6:6 + path_length].decode("utf8"),
                                     raw[6 + path_length + 1:6 + path_length + 1 + content_type_length].decode("utf8"),
                                     SendDataRequestOptions.get(raw[6 + path_length + 2 + content_type_length]),
                                     raw[6 + path_length + 3 + content_type_length:-1])

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        if self.__path is not None:
            ret = utils.int_to_bytes(len(self.__path), num_bytes=1)
            ret += bytearray(self.__path, "utf8")
        else:
            ret = utils.int_to_bytes(0x00, num_bytes=1)
        if self.__content_type is not None:
            ret += utils.int_to_bytes(len(self.__content_type), num_bytes=1)
            ret += bytearray(self.__content_type, "utf8")
        else:
            ret += utils.int_to_bytes(0x00, num_bytes=1)
        ret += utils.int_to_bytes(0x00, num_bytes=1)  # Transport is always TCP
        ret += utils.int_to_bytes(self.__options.code, num_bytes=1)
        if self.__file_data is not None:
            ret += self.__file_data

        return ret

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.PATH_LENGTH:         len(self.__path) if self.__path is not None else 0x00,
                DictKeys.PATH:                self.__path if self.__path is not None else None,
                DictKeys.CONTENT_TYPE_LENGTH: len(self.__content_type) if self.__content_type is not None else 0x00,
                DictKeys.CONTENT_TYPE:        self.__content_type if self.__content_type is not None else None,
                DictKeys.TRANSPORT:           0x00,
                DictKeys.TRANSMIT_OPTIONS:    self.__options,
                DictKeys.RF_DATA:             list(self.__file_data) if self.__file_data is not None else None}

    def __get_path(self):
        """
        Returns the path of the file to upload to Device Cloud.

        Returns:
            String: the path of the file to upload to Device Cloud.
        """
        return self.__path

    def __set_path(self, path):
        """
        Sets the path of the file to upload to Device Cloud.

        Args:
            path (String): the new path of the file to upload to Device Cloud.
        """
        self.__path = path

    def __get_content_type(self):
        """
        Returns the content type of the file to upload.

        Returns:
            String: the content type of the file to upload.
        """
        return self.__content_type

    def __set_content_type(self, content_type):
        """
        Sets the content type of the file to upload.

        Args:
            content_type (String): the new content type of the file to upload.
        """
        self.__content_type = content_type

    def __get_options(self):
        """
        Returns the file upload operation options.

        Returns:
            :class:`.SendDataRequestOptions`: the file upload operation options.

        .. seealso::
           | :class:`.SendDataRequestOptions`
        """
        return self.__options

    def __set_options(self, options):
        """
        Sets the file upload operation options.

        Args:
            options (:class:`.SendDataRequestOptions`): the new file upload operation options

        .. seealso::
           | :class:`.SendDataRequestOptions`
        """
        self.__options = options

    def __get_file_data(self):
        """
        Returns the data of the file to upload.

        Returns:
            Bytearray: the data of the file to upload.
        """
        if self.__file_data is None:
            return None
        return self.__file_data.copy()

    def __set_file_data(self, file_data):
        """
        Sets the data of the file to upload.

        Args:
            file_data (Bytearray): the new data of the file to upload.
        """
        if file_data is None:
            self.__file_data = None
        else:
            self.__file_data = file_data.copy()

    path = property(__get_path, __set_path)
    """String. Path of the file to upload to Device Cloud."""

    content_type = property(__get_content_type, __set_content_type)
    """String. The content type of the file to upload."""

    options = property(__get_options, __set_options)
    """:class:`.SendDataRequestOptions`. File upload operation options."""

    file_data = property(__get_file_data, __set_file_data)
    """Bytearray. Data of the file to upload."""


class SendDataResponsePacket(XBeeAPIPacket):
    """
    This class represents a send data response packet. Packet is built
    using the parameters of the constructor or providing a valid API payload.

    This frame type is sent out the serial port in response to the
    :class:`.SendDataRequestPacket`, providing its frame ID is non-zero.

    .. seealso::
       | :class:`.SendDataRequestPacket`
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 7

    def __init__(self, frame_id, status):
        """
        Class constructor. Instantiates a new :class:`.SendDataResponsePacket` object with the provided parameters.

        Args:
            frame_id (Integer): the frame ID of the packet.
            status (:class:`.DeviceCloudStatus`): the file upload status.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.

        .. seealso::
           | :class:`.DeviceCloudStatus`
           | :class:`.XBeeAPIPacket`
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame id must be between 0 and 255.")

        super().__init__(ApiFrameType.SEND_DATA_RESPONSE)
        self._frame_id = frame_id
        self.__status = status

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Override method.

        Returns:
            :class:`.SendDataResponsePacket`

        Raises:
            InvalidPacketException: if the bytearray length is less than 10. (start delim. + length (2 bytes) + frame
                type + frame id + status + checksum = 7 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.SEND_DATA_RESPONSE`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode.name + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=SendDataResponsePacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.SEND_DATA_RESPONSE.code:
            raise InvalidPacketException("This packet is not a send data response packet.")

        return SendDataResponsePacket(raw[4], DeviceCloudStatus.get(raw[5]))

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return True

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return utils.int_to_bytes(self.__status.code, num_bytes=1)

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.DC_STATUS: self.__status}

    def __get_status(self):
        """
        Returns the file upload status.

        Returns:
            :class:`.DeviceCloudStatus`: the file upload status.

        .. seealso::
           | :class:`.DeviceCloudStatus`
        """
        return self.__status

    def __set_status(self, status):
        """
        Sets the file upload status.

        Args:
            status (:class:`.DeviceCloudStatus`): the new file upload status.

        .. seealso::
           | :class:`.DeviceCloudStatus`
        """
        self.__status = status

    status = property(__get_status, __set_status)
    """:class:`.DeviceCloudStatus`. The file upload status."""
